#pragma once

#include <cmath>

#include<dune/pdelab/common/quadraturerules.hh>
#include<dune/pdelab/finiteelement/localbasiscache.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>



template<typename T, typename X, std::size_t dimR>
class PF_3P_Color_DGF
  : public Dune::PDELab::GridFunctionBase<
      Dune::PDELab::GridFunctionTraits<
        typename T::Traits::GridViewType,
        typename T::template Child<0>::Type::Traits::FiniteElementType
                 ::Traits::LocalBasisType::Traits::RangeFieldType,
        dimR,
        Dune::FieldVector<
          typename T::template Child<0>::Type::Traits::FiniteElementType
                   ::Traits::LocalBasisType::Traits::RangeFieldType,
          dimR
          >
        >,
      PF_3P_Color_DGF<T,X, dimR>
      >
{
  typedef T GFS;

  typedef Dune::PDELab::GridFunctionBase<
    Dune::PDELab::GridFunctionTraits<
      typename T::Traits::GridViewType,
      typename T::template Child<0>::Type::Traits::FiniteElementType
               ::Traits::LocalBasisType::Traits::RangeFieldType,
      dimR,
      Dune::FieldVector<
        typename T::template Child<0>::Type::Traits::FiniteElementType
                 ::Traits::LocalBasisType::Traits::RangeFieldType,
        dimR
        >
      >,
    PF_3P_Color_DGF<T,X,dimR>
    > BaseT;

public:
  typedef typename BaseT::Traits Traits;
  typedef typename T::template Child<0>::Type ChildType;
  typedef typename ChildType::Traits::FiniteElementType
                   ::Traits::LocalBasisType::Traits::RangeFieldType RF;
  typedef typename ChildType::Traits::FiniteElementType
                   ::Traits::LocalBasisType::Traits::RangeType RT;


  PF_3P_Color_DGF(const GFS& gfs, const X& x_)
  : pgfs(stackobject_to_shared_ptr(gfs))
  , lfs(gfs)
  , lfs_cache(lfs)
  , x_view(x_)
  , xl(gfs.maxLocalSize())
  , yb(gfs.maxLocalSize())
  { }

  inline void evaluate (const typename Traits::ElementType& e,
                        const typename Traits::DomainType& x,
                        typename Traits::RangeType& y) const
  {
    lfs.bind(e);
    lfs_cache.update();
    x_view.bind(lfs_cache);
    x_view.read(xl);
    x_view.unbind();

    using namespace Dune::Indices;
    const auto& pf1space = child(lfs,_0);
    const auto& pf2space = child(lfs,_2);

    pf1space.finiteElement().localBasis().
          evaluateFunction(x,yb);
          RF p1 = 0;
          RF p2 = 0;

    for (unsigned int i=0; i<yb.size(); i++)
    {
      p1 += xl[pf1space.localIndex(i)]*yb[i];
      p2 += xl[pf2space.localIndex(i)]*yb[i];
    }
    RF p3 = 1-p1-p2;
    p1=fmax(0,fmin(1,p1));
    p2=fmax(0,fmin(1,p2));
    p3=fmax(0,fmin(1,p3));
    RF pmax=fmax(p1, fmax(p2,p3));
    y[0]=p1/pmax;
    y[1]=p2/pmax;
    y[2]=p3/pmax;
  }

  inline const typename Traits::GridViewType& getGridView () const
  {
    return pgfs->gridView();
  }

private:
  typedef Dune::PDELab::LocalFunctionSpace<GFS> LFS;
  typedef Dune::PDELab::LFSIndexCache<LFS> LFSCache;
  typedef typename X::template ConstLocalView<LFSCache> XView;
  std::shared_ptr<GFS const> pgfs;
  mutable LFS lfs;
  mutable LFSCache lfs_cache;
  mutable XView x_view;
  mutable std::vector<RF> xl;
  mutable std::vector<RT> yb;
};
