
#pragma once
#include <cmath>
#include <dune/pdelab/common/function.hh>
#include <dune/pdelab/constraints/common/constraintsparameters.hh>
#include <dune/phasefield/initial_fn.hh>

#define GRID_EPS_SS 1e-7
#define GRID_WIDTH_SS 1.5      // 1.5 for summer school //2 for contactpointtest //6 for full summer school

#define GRID_BOTTOM_SS 0   //0 For summer school   //0.35 for contactpointtest  //0.35 for hainesjump
#define GRID_TOP_SS 0.75  //0.125 for summer school      //1 for contactpointtest  //0.65 for hainesjump


class v_per : public Dune::PDELab::DirichletConstraintsParameters
{
public:
  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
    if ( xg[1]>GRID_BOTTOM_SS +GRID_EPS_SS  && xg[1]<GRID_TOP_SS -GRID_EPS_SS )
    {
      return false;
    }
    return true;
  }
};



class p_per : public Dune::PDELab::DirichletConstraintsParameters
{
public:
  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
    if ((xg[0]> GRID_WIDTH_SS -GRID_EPS_SS  /*|| xg[0] < GRID_EPS*/) && xg[1]>GRID_BOTTOM_SS +GRID_EPS_SS  && xg[1]<GRID_TOP_SS -GRID_EPS_SS )
    {
      return true;
    }
    return false;
  }
};

class InflowOutflow_u_ss : public Dune::PDELab::DirichletConstraintsParameters
{
public:
  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
    if (xg[0]< GRID_EPS_SS )
    {
      return true;
    }
    return false;
  }
};

class InflowOutflow_phi_ss : public Dune::PDELab::DirichletConstraintsParameters
{
public:
  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
    if ( (xg[0]< GRID_EPS_SS  || xg[0] >GRID_WIDTH_SS -GRID_EPS_SS ) && xg[1]>GRID_BOTTOM_SS +GRID_EPS_SS  && xg[1]<GRID_TOP_SS -GRID_EPS_SS )
    {
      return true;
    }
    return false;
  }
};

template<typename GV, typename RF, typename Params, int dim>
class VelocityInitial_Inflow_ss :
  public Dune::PDELab::AnalyticGridFunctionBase<
  Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim>,
  VelocityInitial_Inflow_ss<GV,RF,Params,dim> >
{
  const Params_Time& time;
  RF meanflow;


public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim> Traits;

  typedef typename Traits::DomainType DomainType;
  typedef typename Traits::RangeType RangeType;

  VelocityInitial_Inflow_ss(const GV & gv, const Params & params) :
    Dune::PDELab::AnalyticGridFunctionBase<Traits,VelocityInitial_Inflow_ss<GV,RF,Params,dim> > (gv)
    , time(params.time)
  {
    meanflow = params.initial.ptree.get("MeanInflow",(RF)0.1);
  }

  template <typename T>
  void setTime(T t){
    time = t;
  }

  inline void evaluateGlobal(const DomainType & x, RangeType & y) const
  {
    y = 0.0;
    if(x[0]<GRID_EPS) //Inflow boundary
    {
      RF factor = meanflow * 6/( (GRID_TOP_SS-GRID_BOTTOM_SS)*(GRID_TOP_SS-GRID_BOTTOM_SS) );
      y[0] = std::max((RF)0, factor * (x[1]-GRID_BOTTOM_SS)*(GRID_TOP_SS-x[1]));
    }
  }
};
